@@ -86,6 +86,30 @@ void bfg_devapi_init()
 }
 
 
+float common_sha256d_and_scrypt_min_nonce_diff(struct cgpu_info * const proc, const struct mining_algorithm * const malgo)
+{
+	switch (malgo->algo)
+	{
+#ifdef USE_SCRYPT
+		case POW_SCRYPT:
+			return 1./0x10000;
+#endif
+#ifdef USE_SHA256D
+		case POW_SHA256D:
+			return 1.;
+#endif
+		default:
+			return -1.;
+	}
+}
+
+#ifdef USE_SCRYPT
+float common_scrypt_min_nonce_diff(struct cgpu_info * const proc, const struct mining_algorithm * const malgo)
+{
+	return (malgo->algo == POW_SCRYPT) ? (1./0x10000) : -1.;
+}
+#endif
+
 bool hashes_done(struct thr_info *thr, int64_t hashes, struct timeval *tvp_hashes, uint32_t *max_nonce)
 {
 	struct cgpu_info *cgpu = thr->cgpu;
@@ -117,22 +141,18 @@ bool hashes_done(struct thr_info *thr, int64_t hashes, struct timeval *tvp_hashe
 	timeradd(&thr->tv_hashes_done, tvp_hashes, &thr->tv_hashes_done);
 	
 	// max_nonce management (optional)
-	if (unlikely((long)thr->tv_hashes_done.tv_sec < cycle)) {
-		int mult;
+	if (max_nonce)
+	{
+		uint64_t new_max_nonce = *max_nonce;
+		new_max_nonce *= cycle;
+		new_max_nonce *= 1000000;
+		new_max_nonce /= ((uint64_t)thr->tv_hashes_done.tv_sec * 1000000) + thr->tv_hashes_done.tv_usec;
 		
-		if (likely(!max_nonce || *max_nonce == 0xffffffff))
-			return true;
+		if (new_max_nonce > 0xffffffff)
+			new_max_nonce = 0xffffffff;
 		
-		mult = 1000000 / ((thr->tv_hashes_done.tv_usec + 0x400) / 0x400) + 0x10;
-		mult *= cycle;
-		if (*max_nonce > (0xffffffff * 0x400) / mult)
-			*max_nonce = 0xffffffff;
-		else
-			*max_nonce = (*max_nonce * mult) / 0x400;
-	} else if (unlikely(thr->tv_hashes_done.tv_sec > cycle) && max_nonce)
-		*max_nonce = *max_nonce * cycle / thr->tv_hashes_done.tv_sec;
-	else if (unlikely(thr->tv_hashes_done.tv_usec > 100000) && max_nonce)
-		*max_nonce = *max_nonce * 0x400 / (((cycle * 1000000) + thr->tv_hashes_done.tv_usec) / (cycle * 1000000 / 0x400));
+		*max_nonce = new_max_nonce;
+	}
 	
 	hashmeter2(thr);
 	
@@ -489,7 +509,7 @@ void cgpu_setup_control_requests(struct cgpu_info * const cgpu)
 {
 	mutex_init(&cgpu->device_mutex);
 	notifier_init(cgpu->thr[0]->mutex_request);
-	pthread_cond_init(&cgpu->device_cond, NULL);
+	pthread_cond_init(&cgpu->device_cond, bfg_condattr);
 }
 
 void cgpu_request_control(struct cgpu_info * const cgpu)
@@ -565,10 +585,15 @@ void minerloop_async(struct thr_info *mythr)
 			}
 			else  // ! should_be_running
 			{
+				if (unlikely(mythr->_job_transition_in_progress && timer_isset(&mythr->tv_morework)))
+				{
+					// Really only happens at startup
+					applog(LOG_DEBUG, "%"PRIpreprv": Job transition in progress, with morework timer enabled: unsetting in-progress flag", proc->proc_repr);
+					mythr->_job_transition_in_progress = false;
+				}
 				if (unlikely((is_running || !mythr->_mt_disable_called) && !mythr->_job_transition_in_progress))
 				{
 disabled: ;
-					timer_unset(&mythr->tv_morework);
 					if (is_running)
 					{
 						if (mythr->busy_state != TBS_GETTING_RESULTS)
@@ -580,6 +605,8 @@ disabled: ;
 					else  // !mythr->_mt_disable_called
 						mt_disable_start__async(mythr);
 				}
+				
+				timer_unset(&mythr->tv_morework);
 			}
 			
 			if (timer_passed(&mythr->tv_morework, &tv_now))
@@ -860,7 +887,7 @@ nohelp:
 	
 	size_t matchlen = 0;
 	if (newvalue)
-		while (!isspace(newvalue[0]))
+		while (newvalue[matchlen] && !isspace(newvalue[matchlen]))
 			++matchlen;
 	
 	for ( ; sdf->optname; ++sdf)
@@ -908,6 +935,7 @@ void _set_auto_sdr(enum bfg_set_device_replytype * const out_success, const char
 		*out_success = SDR_ERR;
 }
 
+static
 const char *_proc_set_device(struct cgpu_info * const proc, const char * const optname, const char * const newvalue, char * const replybuf, enum bfg_set_device_replytype * const out_success)
 {
 	const struct bfg_set_device_definition *sdf;
@@ -936,6 +964,7 @@ const char *_proc_set_device(struct cgpu_info * const proc, const char * const o
 	return replybuf;
 }
 
+static
 const char *__proc_set_device(struct cgpu_info * const proc, char * const optname, char * const newvalue, char * const replybuf, enum bfg_set_device_replytype * const out_success)
 {
 	if (proc->drv->set_device)
@@ -948,8 +977,10 @@ const char *__proc_set_device(struct cgpu_info * const proc, char * const optnam
 	return _proc_set_device(proc, optname, newvalue, replybuf, out_success);
 }
 
-const char *proc_set_device(struct cgpu_info * const proc, char * const optname, char * const newvalue, char * const replybuf, enum bfg_set_device_replytype * const out_success)
+const char *proc_set_device(struct cgpu_info * const proc, char * const optname, char *newvalue, char * const replybuf, enum bfg_set_device_replytype * const out_success)
 {
+	if (!newvalue)
+		newvalue = "";
 	const char * const rv = __proc_set_device(proc, optname, newvalue, replybuf, out_success);
 	switch (*out_success)
 	{
