@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2014 Luke Dashjr
+ * Copyright 2013-2015 Luke Dashjr
  * Copyright 2013-2014 Nate Woolls
  * Copyright 2013 Lingchao Xu
  *
@@ -27,6 +27,7 @@
 #include "util.h"
 
 #define ANTMINER_IO_SPEED 115200
+// ANTMINER_HASH_TIME is for U1/U2 only
 #define ANTMINER_HASH_TIME 0.0000000004761
 
 #define ANTMINER_STATUS_LEN 5
@@ -37,6 +38,8 @@
 #define ANTMINER_COMMAND_OFFSET 32
 
 BFG_REGISTER_DRIVER(antminer_drv)
+static
+const struct bfg_set_device_definition antminer_set_device_funcs[];
 
 static
 bool antminer_detect_one(const char *devpath)
@@ -50,17 +53,21 @@ bool antminer_detect_one(const char *devpath)
 	*info = (struct ICARUS_INFO){
 		.baud = ANTMINER_IO_SPEED,
 		.Hs = ANTMINER_HASH_TIME,
-		.timing_mode = MODE_DEFAULT,
+		.timing_mode = MODE_LONG,
+		.do_icarus_timing = true,
 		.read_size = 5,
+		.reopen_mode = IRM_NEVER,
 	};
 	
-	if (!icarus_detect_custom(devpath, drv, info))
+	struct cgpu_info * const dev = icarus_detect_custom(devpath, drv, info);
+	if (!dev)
 	{
 		free(info);
 		return false;
 	}
 	
-	info->read_count = 15;
+	dev->set_device_funcs = antminer_set_device_funcs;
+	info->read_timeout_ms = 75;
 	
 	return true;
 }
@@ -71,6 +78,8 @@ bool antminer_lowl_probe(const struct lowlevel_device_info * const info)
 	return vcom_lowl_probe_wrapper(info, antminer_detect_one);
 }
 
+// Not used for anything, and needs to read a result for every chip
+#if 0
 static
 char *antminer_get_clock(struct cgpu_info *cgpu, char *replybuf)
 {
@@ -78,6 +87,7 @@ char *antminer_get_clock(struct cgpu_info *cgpu, char *replybuf)
 	unsigned char rebuf[ANTMINER_STATUS_LEN] = {0};
 	
 	struct timeval tv_now;
+	struct timeval tv_timeout, tv_finish;
 	
 	rdreg_buf[0] = 4;
 	rdreg_buf[0] |= 0x80;
@@ -88,7 +98,7 @@ char *antminer_get_clock(struct cgpu_info *cgpu, char *replybuf)
 	applog(LOG_DEBUG, "%"PRIpreprv": Get clock: %02x%02x%02x%02x", cgpu->proc_repr, rdreg_buf[0], rdreg_buf[1], rdreg_buf[2], rdreg_buf[3]);
 	
 	timer_set_now(&tv_now);
-	int err = icarus_write(cgpu->device_fd, rdreg_buf, sizeof(rdreg_buf));
+	int err = icarus_write(cgpu->proc_repr, cgpu->device_fd, rdreg_buf, sizeof(rdreg_buf));
 	
 	if (err != 0)
 	{
@@ -99,7 +109,8 @@ char *antminer_get_clock(struct cgpu_info *cgpu, char *replybuf)
 	applog(LOG_DEBUG, "%"PRIpreprv": Get clock: OK", cgpu->proc_repr);
 	
 	memset(rebuf, 0, sizeof(rebuf));
-	err = icarus_gets(rebuf, cgpu->device_fd, &tv_now, NULL, 10, ANTMINER_STATUS_LEN);
+	timer_set_delay(&tv_timeout, &tv_now, 1000000);
+	err = icarus_read(cgpu->proc_repr, rebuf, cgpu->device_fd, &tv_finish, NULL, &tv_timeout, &tv_now, ANTMINER_STATUS_LEN);
 	
 	// Timeout is ok - checking specifically for an error here
 	if (err == ICA_GETS_ERROR)
@@ -112,9 +123,10 @@ char *antminer_get_clock(struct cgpu_info *cgpu, char *replybuf)
 	
 	return NULL;
 }
+#endif
 
 static
-char *antminer_set_clock(struct cgpu_info *cgpu, char *setting, char *replybuf)
+const char *antminer_set_clock(struct cgpu_info * const cgpu, const char * const optname, const char * const setting, char * const replybuf, enum bfg_set_device_replytype * const out_success)
 {
 	if (!setting || !*setting)
 		return "missing clock setting";
@@ -129,7 +141,7 @@ char *antminer_set_clock(struct cgpu_info *cgpu, char *setting, char *replybuf)
 	}
 	
 	//remove leading character
-	char *hex_setting = setting + 1;
+	const char * const hex_setting = &setting[1];
 
 	uint8_t reg_data[4] = {0};
 	
@@ -149,7 +161,7 @@ char *antminer_set_clock(struct cgpu_info *cgpu, char *setting, char *replybuf)
 	
 	applog(LOG_DEBUG, "%"PRIpreprv": Set clock: %02x%02x%02x%02x", cgpu->proc_repr, cmd_buf[0], cmd_buf[1], cmd_buf[2], cmd_buf[3]);
 	
-	int err = icarus_write(cgpu->device_fd, cmd_buf, sizeof(cmd_buf));
+	int err = icarus_write(cgpu->proc_repr, cgpu->device_fd, cmd_buf, sizeof(cmd_buf));
 		
 	if (err != 0)
 	{
@@ -161,20 +173,55 @@ char *antminer_set_clock(struct cgpu_info *cgpu, char *setting, char *replybuf)
 	
 	// This is confirmed required in order for the clock change to "take"
 	cgsleep_ms(500);
-		
-	return antminer_get_clock(cgpu, replybuf);
+	
+	return NULL;
 }
 
 static
-char *antminer_set_device(struct cgpu_info *cgpu, char *option, char *setting, char *replybuf)
-{		
-	if (strcasecmp(option, "clock") == 0)
+const char *antminer_set_voltage(struct cgpu_info * const proc, const char * const optname, const char * const newvalue, char * const replybuf, enum bfg_set_device_replytype * const out_success)
+{
+	if (!(newvalue && *newvalue))
+		return "Missing voltage value";
+	
+	// For now we only allow hex values that use BITMAINtech's lookup table
+	// This means values should be prefixed with an x so that later we can
+	// accept and distinguish decimal values
+	if (newvalue[0] != 'x' || strlen(newvalue) != 4)
+invalid_voltage:
+		return "Only raw voltage configurations are currently supported using 'x' followed by 3 hexadecimal digits";
+	
+	char voltagecfg_hex[5];
+	voltagecfg_hex[0] = '0';
+	memcpy(&voltagecfg_hex[1], &newvalue[1], 3);
+	voltagecfg_hex[4] = '\0';
+	
+	uint8_t cmd[4];
+	if (!hex2bin(&cmd[1], voltagecfg_hex, 2))
+		goto invalid_voltage;
+	cmd[0] = 0xaa;
+	cmd[1] |= 0xb0;
+	cmd[3] = 0;
+	cmd[3] = crc5usb(cmd, (4 * 8) - 5);
+	cmd[3] |= 0xc0;
+	
+	if (opt_debug)
 	{
-		return antminer_set_clock(cgpu, setting, replybuf);
+		char hex[(4 * 2) + 1];
+		bin2hex(hex, cmd, 4);
+		applog(LOG_DEBUG, "%"PRIpreprv": Set voltage: %s", proc->proc_repr, hex);
 	}
-
-	sprintf(replybuf, "Unknown option: %s", option);
-	return replybuf;
+	
+	const int err = icarus_write(proc->proc_repr, proc->device_fd, cmd, sizeof(cmd));
+	
+	if (err)
+	{
+		sprintf(replybuf, "Error sending set voltage (err=%d)", err);
+		return replybuf;
+	}
+	
+	applog(LOG_DEBUG, "%"PRIpreprv": Set voltage: OK", proc->proc_repr);
+	
+	return NULL;
 }
 
 static
@@ -191,7 +238,7 @@ void antminer_flash_led(const struct cgpu_info *antminer)
 	cmd_buf[offset + 3] = crc5usb(cmd_buf, sizeof(cmd_buf));
 
 	const int fd = antminer->device_fd;
-	icarus_write(fd, (char *)(&cmd_buf), sizeof(cmd_buf));
+	icarus_write(antminer->proc_repr, fd, (char *)(&cmd_buf), sizeof(cmd_buf));
 }
 
 static
@@ -206,6 +253,17 @@ bool antminer_identify(struct cgpu_info *antminer)
 	return true;
 }
 
+static
+const struct bfg_set_device_definition antminer_set_device_funcs[] = {
+	{"baud"         , icarus_set_baud         , "serial baud rate"},
+	{"work_division", icarus_set_work_division, "number of pieces work is split into"},
+	{"reopen"       , icarus_set_reopen       , "how often to reopen device: never, timeout, cycle, (or now for a one-shot reopen)"},
+	{"timing"       , icarus_set_timing       , "timing of device; see README.FPGA"},
+	{"clock", antminer_set_clock, "clock frequency"},
+	{"voltage", antminer_set_voltage, "voltage ('x' followed by 3 digit hex code)"},
+	{NULL},
+};
+
 static
 void antminer_drv_init()
 {
@@ -213,7 +271,6 @@ void antminer_drv_init()
 	antminer_drv.dname = "antminer";
 	antminer_drv.name = "AMU";
 	antminer_drv.lowl_probe = antminer_lowl_probe;
-	antminer_drv.set_device = antminer_set_device,
 	antminer_drv.identify_device = antminer_identify;
 	++antminer_drv.probe_priority;
 }
