@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2014 Luke Dashjr
+ * Copyright 2013-2015 Luke Dashjr
  * Copyright 2012-2014 Con Kolivas
  * Copyright 2011 Andrew Smith
  * Copyright 2011 Jeff Garzik
@@ -106,6 +106,10 @@ const char *bfg_json_obj_string(json_t *json, const char *key, const char *fail)
 
 extern const char *__json_array_string(json_t *, unsigned int entry);
 
+#ifndef min
+#  define min(a, b)  ((a) < (b) ? (a) : (b))
+#endif
+
 extern void *my_memrchr(const void *, int, size_t);
 
 extern bool isCalpha(int);
@@ -121,6 +125,8 @@ bool isCspace(int c)
 	}
 }
 
+extern bool match_strtok(const char *optlist, const char *delim, const char *needle);
+
 typedef bool (*appdata_file_callback_t)(const char *, void *);
 extern bool appdata_file_call(const char *appname, const char *filename, appdata_file_callback_t, void *userp);
 extern char *appdata_file_find_first(const char *appname, const char *filename);
@@ -188,7 +194,6 @@ extern char *ucs2_to_utf8_dup(uint16_t *in, size_t sz);
 }while(0)
 
 extern void gen_hash(unsigned char *data, unsigned char *hash, int len);
-extern void hash_data(unsigned char *out_hash, const unsigned char *data);
 extern void real_block_target(unsigned char *target, const unsigned char *data);
 extern bool hash_target_check(const unsigned char *hash, const unsigned char *target);
 extern bool hash_target_check_v(const unsigned char *hash, const unsigned char *target);
@@ -220,6 +225,10 @@ int cgtimer_to_ms(cgtimer_t *cgt)
 	return (cgt->tv_sec * 1000) + (cgt->tv_usec / 1000);
 }
 
+extern int bfg_cond_timedwait(pthread_cond_t * restrict, pthread_mutex_t * restrict, const struct timeval *);
+extern pthread_condattr_t *bfg_condattr_();
+#define bfg_condattr (bfg_condattr_())
+
 #define cgtimer_sub(a, b, res)  timersub(a, b, res)
 double us_tdiff(struct timeval *end, struct timeval *start);
 double tdiff(struct timeval *end, struct timeval *start);
@@ -253,6 +262,9 @@ typedef SOCKETTYPE notifier_t[2];
 extern void notifier_init(notifier_t);
 extern void notifier_wake(notifier_t);
 extern void notifier_read(notifier_t);
+extern bool notifier_wait(notifier_t, const struct timeval *);
+extern bool notifier_wait_us(notifier_t, unsigned long long usecs);
+extern void notifier_reset(notifier_t);
 extern void notifier_init_invalid(notifier_t);
 extern void notifier_destroy(notifier_t);
 
@@ -454,7 +466,7 @@ void pk_u64le(void * const bufp, const int offset, const uint64_t nv)
 }while(0)
 
 #define is_power_of_two(n)  \
-	(0 == ((n) && ((n) - 1)))
+	(0 == ((n) & ((n) - 1)))
 
 static inline
 uint32_t upper_power_of_two_u32(uint32_t n)
@@ -631,7 +643,7 @@ static inline
 void bytes_free(bytes_t *b)
 {
 	free(b->buf);
-	b->sz = b->allocsz = 0;
+	bytes_init(b);
 }
 
 
@@ -711,6 +723,15 @@ int timer_elapsed(const struct timeval *tvp_timer, const struct timeval *tvp_now
 	return tv.tv_sec;
 }
 
+static inline
+long timer_remaining_us(const struct timeval *tvp_timer, const struct timeval *tvp_now)
+{
+	struct timeval tv;
+	const struct timeval *_tvp_now = _bfg_nullisnow(tvp_now, &tv);
+	timersub(tvp_timer, _tvp_now, &tv);
+	return timeval_to_us(&tv);
+}
+
 static inline
 bool timer_passed(const struct timeval *tvp_timer, const struct timeval *tvp_now)
 {
